import numpy as np

from mountainlab_pytools import mdaio
import json

processor_name='ephys.compare_ground_truth'
processor_version='0.1'

def create_occupancy_array(times,labels,segment_size,num_segments,K,*,spread,multiplicity):
    ret=np.zeros((K,num_segments),dtype=np.int)
    counts=np.zeros(K,dtype=np.int64)
    for k in range(1,K+1):
        inds_k=np.where(labels==k)[0]
        times_k=times[inds_k]
        labels_k=labels[inds_k]
        segment_inds=np.floor(times_k/segment_size).astype(np.int64)
        tmp=np.bincount(segment_inds)
        if not multiplicity:
            tmp[np.where(ret>=1)[0]]=1
        tmp=tmp[0:num_segments] # make sure it does not overstep the bounds!
        ret[k-1,0:len(tmp)]=tmp
        if spread:
            ret[k-1,np.maximum(0,segment_inds-1)]=1
            ret[k-1,np.minimum(num_segments-1,segment_inds+1)]=1
        counts[k-1]=len(inds_k)
    return (ret,counts)

def compare_ground_truth(*,firings_true,firings,json_out,max_dt=20):
    """
    compare a sorting (firings) with ground truth (firings_true)

    Parameters
    ----------
    firings_true : INPUT
        Path of true firings file (RxL) R>=3, L=#evts
    firings : INPUT
        Path of sorted firings file (RxL) R>=3, L=#evts
    json_out : OUTPUT
        Path of the output file containing the results of the comparison
        
    max_dt : int
        Tolerance for matching events (in timepoints)
        
    """

    print('Reading arrays...')
    F=mdaio.readmda(firings)
    Ft=mdaio.readmda(firings_true)
    print('Initializing data...')
    L=F.shape[1]
    Lt=Ft.shape[1]
    times=F[1,:]
    times_true=Ft[1,:]
    labels=F[2,:].astype(np.int32)
    labels_true=Ft[2,:].astype(np.int32)
    F=0  # free memory? Alex: depends if python decides to call garbage collection
    Ft=0 # free memory?
    N=np.maximum(np.max(times),np.max(times_true))
    K=np.max(labels)
    Kt=np.max(labels_true)

    # todo: subsample in first pass to get the best
    
    # First we split into segments
    print('Splitting into segments')
    segment_size=max_dt
    num_segments=int(np.ceil(N/segment_size))
    N=num_segments*segment_size
    segments=[]
    # occupancy: K x num_segments
    occupancy,counts=create_occupancy_array(times,labels,segment_size,num_segments,K,spread=False,multiplicity=True)
    # occupancy_true: Kt x num_segments
    occupancy_true,counts_true=create_occupancy_array(times_true,labels_true,segment_size,num_segments,Kt,spread=True,multiplicity=False)
    
    # Note: we spread the occupancy_true but not the occupancy
    # Note: we count the occupancy with multiplicity but not occupancy_true
    
    print('Computing pairwise counts and accuracies...')
    pairwise_counts=occupancy_true @ occupancy.transpose() # Kt x K
    pairwise_accuracies=np.zeros((Kt,K))
    for k1 in range(1,Kt+1):
        for k2 in range(1,K):
            numer=pairwise_counts[k1-1,k2-1]
            denom=counts_true[k1-1]+counts[k2-1]-numer
            if denom>0:
                pairwise_accuracies[k1-1,k2-1]=numer/denom
    
    print('Preparing output...')
    ret={
        "true_units":{}
    }
    for k1 in range(1,Kt+1):
        k2_match=int(1+np.argmax(pairwise_accuracies[k1-1,:].ravel()))
        # todo: compute accuracy more precisely here
        num_matches=int(pairwise_counts[k1-1,k2_match-1])
        num_false_positives=int(counts[k2_match-1]-num_matches)
        num_false_negatives=int(counts_true[k1-1]-num_matches)
        unit={
            "best_match":k2_match,
            "accuracy":pairwise_accuracies[k1-1,k2_match-1],
            "num_matches":num_matches,
            "num_false_positives":num_false_positives,
            "num_false_negatives":num_false_negatives
        }
        ret['true_units'][k1]=unit
        
    print('Writing output...')
    str=json.dumps(ret,indent=4)
    with open(json_out, 'w') as out:
        out.write(str)
    print('Done.')
    
    return True

compare_ground_truth.name=processor_name
compare_ground_truth.version=processor_version



import numpy as np
import scipy.interpolate as spi

import sys,os
parent_path=os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(parent_path)

from p_synthesize_timeseries import synthesize_timeseries

from mountainlab_pytools import mdaio

processor_name='pyms.synthesize_drifting_timeseries'
processor_version='0.1'
def synthesize_drifting_timeseries(*,
        firings,
        waveforms,
        timeseries_out=None,
        noise_level=1,
        samplerate=30000,
        duration=60,
        waveform_upsamplefac=1,
        amplitudes_row=0,
        num_interp_nodes=2
    ):
    """
    Synthesize a electrophysiology timeseries from a set of ground-truth firing events and waveforms, and simulating drift (linear for now)

    Parameters
    ----------
    firings : INPUT
        (Optional) The path of firing events file in .mda format. RxL where 
        R>=3 and L is the number of events. Second row is the timestamps, 
        third row is the integer labels
    waveforms : INPUT
        (Optional) The path of (possibly upsampled) waveforms file in .mda
        format. Mx(T*waveform_upsample_factor)*(2K), where M is the number of
        channels, T is the clip size, and K is the number of units. Each unit
        has a contiguous pair of waveforms (interpolates from first to second
        across the timeseries)
    
    timeseries_out : OUTPUT
        The output path for the new timeseries. MxN

    noise_level : double
        (Optional) Standard deviation of the simulated background noise added to the timeseries
    samplerate : double
        (Optional) Sample rate for the synthetic dataset in Hz
    duration : double
        (Optional) Duration of the synthetic dataset in seconds. The number of timepoints will be duration*samplerate
    waveform_upsamplefac : int
        (Optional) The upsampling factor corresponding to the input waveforms. (avoids digitization artifacts)
    amplitudes_row : int
        (Optional) If positive, this is the row in the firings arrays where the amplitude scale factors are found. Otherwise, use all 1's
    num_interp_nodes : int
        (Optional) For drift, the number of timepoints where we specify the waveform (Default 2)
    """
    
    if type(firings)==str:
        F=mdaio.readmda(firings)
    else:
        F=firings

    if amplitudes_row==0:
        F=np.concatenate((F,np.ones((1,F.shape[1]))))
        amplitudes_row=F.shape[0]
        
    times=F[1,:]
    times_normalized=times/(duration*samplerate) #normalized between 0 and 1
    labels=F[2,:]
    amps=F[amplitudes_row-1,:]
            
    F=np.kron(F,[1]*num_interp_nodes) #duplicate every event!
    
    for j in range(num_interp_nodes):
        F[amplitudes_row-1,j::num_interp_nodes]=amps*time_basis_func(j,num_interp_nodes,times_normalized)
        # adjust the labels
        F[2,j::num_interp_nodes]=(labels-1)*num_interp_nodes+j+1 #remember that labels are 1-indexed
    return synthesize_timeseries(
        firings=F,
        waveforms=waveforms,
        timeseries_out=timeseries_out,
        noise_level=noise_level,
        samplerate=samplerate,
        duration=duration,
        waveform_upsamplefac=waveform_upsamplefac,
        amplitudes_row=amplitudes_row
    )
    
def time_basis_func(j,nnodes,times):
    if nnodes==1:
        return np.ones(times.shape)
    v=np.zeros(nnodes)
    v[j]=1
    tnodes=np.arange(nnodes)/(nnodes-1)
    if nnodes<3:
        kind0='linear'
    elif nnodes==3:
        kind0='quadratic'
    else:
        kind0='cubic'
    return spi.interp1d(tnodes,v,kind=kind0)(times)

def test_synthesize_drifting_timeseries():
    waveform_upsamplefac=5
    M=4 #num channels
    T=800 #num timepoints per clip
    K=2 # num units
    L=20
    num_interp_nodes=4
    from p_synthesize_random_waveforms import synthesize_random_waveforms
    #W=np.random.rand(M,T*waveform_upsamplefac,K*2)
    (W,geom)=synthesize_random_waveforms(M=M,T=T,K=K*num_interp_nodes,upsamplefac=waveform_upsamplefac)
    
    times=np.arange(L)/(L-1)*10000
    #labels=np.tile([1,2],int(L/2))
    labels=np.tile([1],L)
    F=np.array([np.zeros(L),times,labels])
    #return;
    X=synthesize_drifting_timeseries(waveforms=W,firings=F,duration=1,samplerate=10000,noise_level=0,num_interp_nodes=num_interp_nodes)
    import matplotlib.pyplot as pp
    pp.imshow(X,extent=[0, 1, 0, 1])
    #mdaio.writemda32(X,'tmp.mda');
    return True

synthesize_drifting_timeseries.test=test_synthesize_drifting_timeseries
synthesize_drifting_timeseries.name = processor_name
synthesize_drifting_timeseries.version = processor_version

if __name__ == '__main__':
    print ('Running test')
    test_synthesize_drifting_timeseries()



import numpy as np

from mountainlab_pytools import mdaio

processor_name='ephys.synthesize_timeseries'
processor_version='0.11a'
def synthesize_timeseries(*,firings='',waveforms='',timeseries_out,noise_level=1,samplerate=30000,duration=60,waveform_upsamplefac,amplitudes_row=0):
    """
    Synthesize an electrophysiology timeseries from a set of ground-truth firing events and waveforms

    Parameters
    ----------
    firings : INPUT
        (Optional) The path of firing events file in .mda format. RxL where R>=3 and L is the number of events. Second row is the timestamps, third row is the integer labels/
    waveforms : INPUT
        (Optional) The path of (possibly upsampled) waveforms file in .mda format. Mx(T*waveform_upsample_factor)*K, where M is the number of channels, T is the clip size, and K is the number of units.
    
    timeseries_out : OUTPUT
        The output path for the new timeseries. MxN

    noise_level : double
        (Optional) Standard deviation of the simulated background noise added to the timeseries
    samplerate : double
        (Optional) Sample rate for the synthetic dataset in Hz
    duration : double
        (Optional) Duration of the synthetic dataset in seconds. The number of timepoints will be duration*samplerate
    waveform_upsamplefac : int
        (Optional) The upsampling factor corresponding to the input waveforms. (avoids digitization artifacts)
    amplitudes_row : int
        (Optional) If positive, this is the row in the firings arrays where the amplitude scale factors are found. Otherwise, use all 1's
    """
    num_timepoints=np.int64(samplerate*duration)
    waveform_upsamplefac=int(waveform_upsamplefac)
    
    if type(waveforms)==str:
        if waveforms:
            W=mdaio.readmda(waveforms)
        else:
            W=np.zeros((4,100*waveform_upsamplefac,0))
    else:
        W=waveforms
        
    
    
    if type(firings)==str:
        if firings:
            F=mdaio.readmda(firings)
        else:
            F=np.zeros((3,0))
    else:
        F=firings
        
        
    times=F[1,:]
    labels=F[2,:].astype('int')
    
    M,TT,K = W.shape[0],W.shape[1],W.shape[2]
    T=int(TT/waveform_upsamplefac)
    Tmid=int(np.ceil((T+1)/2-1))
    
    N=num_timepoints
    if (N==0):
        if times.size==0:
            N=T
        else:
            N=max(times)+T
            
    X=np.random.randn(M,N)*noise_level

    waveform_list=[]
    for k in range(K):
        waveform0=W[:,:,k-1]
        waveform_list.append(waveform0)

    for j in range(times.size):
        t0=times[j]
        k0=labels[j]
        amp0=1
        if amplitudes_row>0:
            amp0=F[amplitudes_row-1,j]
        waveform0=waveform_list[k0-1]
        frac_offset=int(np.floor((t0-np.floor(t0))*waveform_upsamplefac))
        tstart=np.int64(np.floor(t0))-Tmid
        if (0<=tstart) and (tstart+T<=N):
            X[:,tstart:tstart+T]=X[:,tstart:tstart+T]+waveform0[:,frac_offset::waveform_upsamplefac]*amp0

    if timeseries_out:
        return mdaio.writemda32(X,timeseries_out)
    else:
        return (X)

def test_synthesize_timeseries():
    #ret=synthesize_timeseries(timeseries_out='tmp.mda',duration=60,samplerate=30000)
    waveform_upsamplefac=13
    W=np.random.rand(4,100*waveform_upsamplefac,6)  
    F=np.zeros((3,0))
    X=synthesize_timeseries(waveforms=W,firings=F,duration=60,samplerate=30000,amplitudes_row=4)
    print(X.shape)
    #import matplotlib.pyplot as pp
    #pp.imshow(X,extent=[0, 5, 0, .3]);
    #mdaio.writemda32(X,'tmp.mda');
    return True

synthesize_timeseries.test=test_synthesize_timeseries
synthesize_timeseries.name = processor_name
synthesize_timeseries.version = processor_version

if __name__ == '__main__':
    print ('Running test')
    test_synthesize_timeseries()



import numpy as np
import h5py
import deepdish as dd

from mountainlab_pytools import mdaio
from timeserieschunkreader import TimeseriesChunkReader

processor_name='ephys.compute_cross_correlograms'
processor_version='0.11'
def compute_cross_correlograms(*,firings,correlograms_out,samplerate,mode='autocorrelograms',max_dt_msec=50,bin_size_msec=2):
    """
    Compute cross-correlograms for a firings file

    Parameters
    ----------
    firings : INPUT
        Path of firings mda file (RxL) where R>=3 and L is the number of events. Second row are timestamps, third row are integer labels.    
        
    correlograms_out : OUTPUT
        Path of output hdf5 file
    
    mode : string
        Choices: autocorrelograms, (more to come)    
    samplerate : float
        Sample rate in Hz
    max_dt_msec : float
        Max dt for the histograms (msec)
    bin_size_msec : float
        Bin size for the histograms (msec)
    """    
    X=compute_cross_correlograms_helper(firings=firings,mode=mode,samplerate=samplerate,max_dt_msec=max_dt_msec,bin_size_msec=bin_size_msec)
    dd.io.save(correlograms_out,X)
    return True
    
# Same as compute_cross_correlograms, except return the result in memory
def compute_cross_correlograms_helper(*,firings,mode='autocorrelograms',samplerate=30000,max_dt_msec=50,bin_size_msec=2):
    if type(firings)==str:
        F=mdaio.readmda(firings)
    else:
        F=firings
    R,L=np.shape(F)
    assert(R>=3)
    assert(mode=='autocorrelograms')
    max_dt_tp=max_dt_msec/1000*samplerate
    bin_size_tp=bin_size_msec/1000*samplerate
    times=F[1,:]
    labels=F[2,:].astype(int)
    K=np.max(labels)
    correlograms=[]
    for k in range(1,K+1):
        inds_k=np.where(labels==k)[0]
        times_k=times[inds_k]
        bin_counts,bin_edges=compute_autocorrelogram(times_k,max_dt_tp=max_dt_tp,bin_size_tp=bin_size_tp)
        correlograms.append({
            "k":k,
            "bin_edges":bin_edges/samplerate*1000,
            "bin_counts":bin_counts
        })
    return {
        "correlograms":correlograms
    }

def compute_autocorrelogram(times,*,max_dt_tp,bin_size_tp):
    num_bins_left=int(max_dt_tp/bin_size_tp)
    L=len(times)
    times2=np.sort(times)
    step=1
    candidate_inds=np.arange(L)
    vals_list=[]
    while True:
        candidate_inds=candidate_inds[candidate_inds+step<L]
        candidate_inds=candidate_inds[times2[candidate_inds+step]-times2[candidate_inds]<=max_dt_tp]
        if len(candidate_inds)>0:
            vals=times2[candidate_inds+step]-times2[candidate_inds]
            vals_list.append(vals)
            vals_list.append(-vals)
        else:
            break
        step+=1
    if len(vals_list)>0:
        all_vals=np.concatenate(vals_list)
    else:
        all_vals=np.array([]);
    aa=np.arange(-num_bins_left,num_bins_left+1)*bin_size_tp
    all_vals=np.sign(all_vals)*(np.abs(all_vals)-bin_size_tp*0.00001) # a trick to make the histogram symmetric
    bin_counts,bin_edges=np.histogram(all_vals,bins=aa)
    return (bin_counts,bin_edges)
    
compute_cross_correlograms.name=processor_name
compute_cross_correlograms.version=processor_version



import numpy as np

from mountainlab_pytools import mdaio
from timeserieschunkreader import TimeseriesChunkReader

processor_name='ephys.compute_templates'
processor_version='0.11'
def compute_templates(*,timeseries,firings,templates_out,clip_size=100):
    """
    Compute templates (average waveforms) for clusters defined by the labeled events in firings.

    Parameters
    ----------
    timeseries : INPUT
        Path of timeseries mda file (MxN) from which to draw the event clips (snippets) for computing the templates. M is number of channels, N is number of timepoints.
    firings : INPUT
        Path of firings mda file (RxL) where R>=3 and L is the number of events. Second row are timestamps, third row are integer labels.    
        
    templates_out : OUTPUT
        Path of output mda file (MxTxK). T=clip_size, K=maximum cluster label. Note that empty clusters will correspond to a template of all zeros. 
        
    clip_size : int
        (Optional) clip size, aka snippet size, number of timepoints in a single template
    """    
    templates=compute_templates_helper(timeseries=timeseries,firings=firings,clip_size=clip_size)
    return mdaio.writemda32(templates,templates_out)
    
# Same as compute_templates, except return the templates as an array in memory
def compute_templates_helper(*,timeseries,firings,clip_size=100):
    X=mdaio.DiskReadMda(timeseries)
    M,N = X.N1(),X.N2()
    F=mdaio.readmda(firings)
    L=F.shape[1]
    L=L
    T=clip_size
    Tmid = int(np.floor((T + 1) / 2) - 1);
    times=F[1,:].ravel()
    labels=F[2,:].ravel().astype(int)
    K=np.max(labels)

    sums=np.zeros((M,T,K),dtype='float64')
    counts=np.zeros(K)

    for k in range(1,K+1):
        inds_k=np.where(labels==k)[0]
        #TODO: subsample
        for ind_k in inds_k:
            t0=int(times[ind_k])
            if (clip_size<=t0) and (t0<N-clip_size):
                clip0=X.readChunk(i1=0,N1=M,i2=t0-Tmid,N2=T)
                sums[:,:,k-1]+=clip0
                counts[k-1]+=1
    templates=np.zeros((M,T,K))
    for k in range(K):
        templates[:,:,k]=sums[:,:,k]/counts[k]
    return templates
    
compute_templates.name=processor_name
compute_templates.version=processor_version
def test_compute_templates():
    M,N,K,T,L = 5,1000,6,50,100
    X=np.random.rand(M,N)
    mdaio.writemda32(X,'tmp.mda')
    F=np.zeros((3,L))
    F[1,:]=1+np.random.randint(N,size=(1,L))
    F[2,:]=1+np.random.randint(K,size=(1,L))
    mdaio.writemda64(F,'tmp2.mda')
    ret=compute_templates(timeseries='tmp.mda',firings='tmp2.mda',templates_out='tmp3.mda',clip_size=T)
    assert(ret)
    templates0=mdaio.readmda('tmp3.mda')
    assert(templates0.shape==(M,T,K))
    return True
compute_templates.test=test_compute_templates



import numpy as np

from mountainlab_pytools import mdaio
from p_compute_templates import compute_templates_helper
from p_compute_cross_correlograms import compute_cross_correlograms_helper
import json

processor_name='ephys.compute_cluster_metrics'
processor_version='0.11'
def compute_cluster_metrics(*,timeseries='',firings,metrics_out,clip_size=100,samplerate=0,
        refrac_msec=1):
    """
    Compute cluster metrics for a spike sorting output

    Parameters
    ----------
    firings : INPUT
        Path of firings mda file (RxL) where R>=3 and L is the number of events. Second row are timestamps, third row are integer cluster labels.
    timeseries : INPUT
        Optional path of timeseries mda file (MxN) which could be raw or preprocessed   
        
    metrics_out : OUTPUT
        Path of output json file containing the metrics.
        
    clip_size : int
        (Optional) clip size, aka snippet size (used when computing the templates, or average waveforms)
    samplerate : float
        Optional sample rate in Hz

    refrac_msec : float
        (Optional) Define interval (in ms) as refractory period. If 0 then don't compute the refractory period metric.
    """    
    print('Reading firings...')
    F=mdaio.readmda(firings)

    print('Initializing...')
    R=F.shape[0]
    L=F.shape[1]
    assert(R>=3)
    times=F[1,:]
    labels=F[2,:].astype(np.int)
    K=np.max(labels)
    N=0
    if timeseries:
        X=mdaio.DiskReadMda(timeseries)
        N=X.N2()

    if (samplerate>0) and (N>0):
        duration_sec=N/samplerate
    else:
        duration_sec=0

    clusters=[]
    for k in range(1,K+1):
        inds_k=np.where(labels==k)[0]
        metrics_k={
            "num_events":len(inds_k)
        }
        if duration_sec:
            metrics_k['firing_rate']=len(inds_k)/duration_sec
        cluster={
            "label":k,
            "metrics":metrics_k
        }
        clusters.append(cluster)

    if timeseries:
        print('Computing templates...')
        templates=compute_templates_helper(timeseries=timeseries,firings=firings,clip_size=clip_size)
        for k in range(1,K+1):
            template_k=templates[:,:,k-1]
            # subtract mean on each channel (todo: vectorize this operation)
            # Alex: like this?
            # template_k[m,:] -= np.mean(template_k[m,:])
            for m in range(templates.shape[0]):
                template_k[m,:]=template_k[m,:]-np.mean(template_k[m,:])
            peak_amplitude=np.max(np.abs(template_k))
            clusters[k-1]['metrics']['peak_amplitude']=peak_amplitude
        ## todo: subtract template means, compute peak amplitudes

    if refrac_msec > 0:
        print('Computing Refractory Period Violations')
        msec_string = '{0:.5f}'.format(refrac_msec).rstrip('0').rstrip('.')
        rr_string = 'refractory_violation_{}msec'.format(msec_string)
        max_dt_msec=50
        bin_size_msec=refrac_msec
        auto_cors = compute_cross_correlograms_helper(firings=firings,mode='autocorrelograms',samplerate=samplerate,max_dt_msec=max_dt_msec,bin_size_msec=bin_size_msec)
        mid_ind  = np.floor(max_dt_msec/bin_size_msec).astype(int)
        mid_inds = np.arange(mid_ind-2,mid_ind+2)
        for k0,auto_cor_obj in enumerate(auto_cors['correlograms']):
            auto = auto_cor_obj['bin_counts']
            k    = auto_cor_obj['k']
            peak = np.percentile(auto, 75)
            mid  = np.mean(auto[mid_inds])
            rr   = safe_div(mid,peak)
            clusters[k-1]['metrics'][rr_string] = rr

    ret={
        "clusters":clusters
    }

    print('Writing output...')
    str=json.dumps(ret,indent=4)
    with open(metrics_out, 'w') as out:
        out.write(str)
    print('Done.')
    return True

def safe_div(x,y):
    if y == 0:
        return 0
    return x / y

compute_cluster_metrics.name=processor_name
compute_cluster_metrics.version=processor_version



import sys

from mountainlab_pytools import processormanager as pm

import p_bandpass_filter
import p_whiten

PM=pm.ProcessorManager()

PM.registerProcessor(p_bandpass_filter.bandpass_filter)
PM.registerProcessor(p_whiten.whiten)

if not PM.run(sys.argv):
    exit(-1)



import numpy as np

from mountainlab_pytools import mdaio

processor_name='ephys.synthesize_random_firings'
processor_version='0.14'
def synthesize_random_firings(*,firings_out,K=20,samplerate=30000,duration=60):
    """
    Synthesize random waveforms for use in creating a synthetic timeseries dataset

    Parameters
    ----------
    firings_out : OUTPUT
        Path to output firings mda file. 3xL, L is the number of events, second row are timestamps, third row are integer unit labels
    
    K : int
        (Optional) number of simulated units
    samplerate : double
        (Optional) sampling frequency in Hz
    duration : double
        (Optional) duration of the simulated acquisition in seconds
    """
    firing_rates=3*np.ones((K))
    refr=4
    
    N=np.int64(duration*samplerate)
    
    # events/sec * sec/timepoint * N
    populations=np.ceil(firing_rates/samplerate*N).astype('int')
    times=np.zeros(0)
    labels=np.zeros(0)
    for k in range(1,K+1):
        refr_timepoints=refr/1000*samplerate
        
        times0=np.random.rand(populations[k-1])*(N-1)+1

        ## make an interesting autocorrelogram shape
        times0=np.hstack((times0,times0+rand_distr2(refr_timepoints,refr_timepoints*20,times0.size)))
        times0=times0[np.random.choice(times0.size,int(times0.size/2))]
        times0=times0[np.where((0<=times0)&(times0<N))]
        
        times0=enforce_refractory_period(times0,refr_timepoints)
        times=np.hstack((times,times0))
        labels=np.hstack((labels,k*np.ones(times0.shape)))

    sort_inds=np.argsort(times)
    times=times[sort_inds]
    labels=labels[sort_inds]

    firings=np.zeros((3,times.size),dtype=np.float64)
    firings[1,:]=times
    firings[2,:]=labels
    return mdaio.writemda64(firings,firings_out)
    
def rand_distr2(a,b,num):
    X=np.random.rand(num)
    X=a+(b-a)*X**2
    return X
    
def enforce_refractory_period(times_in,refr):
    if (times_in.size==0): return times_in
    
    times0=np.sort(times_in)
    done=False
    while not done:
        diffs=times0[1:]-times0[:-1]
        diffs=np.hstack((diffs,np.inf)) #hack to make sure we handle the last one
        inds0=np.where((diffs[:-1]<=refr)&(diffs[1:]>=refr))[0] #only first violator in every group
        if len(inds0)>0:
            times0[inds0]=-1 #kind of a hack, what's the better way?
            times0=times0[np.where(times0>=0)]
        else:
            done=True
        
    return times0
        
def test_synthesize_random_firings():
    K=10
    synthesize_random_firings(K=K,firings_out='tmp.firings.mda')
    firings=mdaio.readmda('tmp.firings.mda')
    labels=firings[2,:]
    assert(max(labels)==K)
    assert(firings.shape[0]==3)    
    return True

synthesize_random_firings.test=test_synthesize_random_firings
synthesize_random_firings.name = processor_name
synthesize_random_firings.version = processor_version

if __name__ == '__main__':
    print ('Running test')
    test_synthesize_random_firings()
    



import sys

from mountainlab_pytools import processormanager as pm

import p_compare_ground_truth

PM=pm.ProcessorManager()

PM.registerProcessor(p_compare_ground_truth.compare_ground_truth)

if not PM.run(sys.argv):
    exit(-1)



import sys

from mountainlab_pytools import processormanager as pm

import p_compute_templates
import p_compute_cross_correlograms
import p_convert_array
import p_compute_cluster_metrics

PM=pm.ProcessorManager()

PM.registerProcessor(p_compute_templates.compute_templates)
PM.registerProcessor(p_compute_cross_correlograms.compute_cross_correlograms)
PM.registerProcessor(p_convert_array.convert_array)
PM.registerProcessor(p_compute_cluster_metrics.compute_cluster_metrics)

if not PM.run(sys.argv):
    exit(-1)



import numpy as np

def exp_growth(amp1,amp2,dur1,dur2):
    t=np.arange(0,dur1)
    Y=np.exp(t/dur2)
    # Want Y[0]=amp1
    # Want Y[-1]=amp2
    Y=Y/(Y[-1]-Y[0])*(amp2-amp1)
    Y=Y-Y[0]+amp1;
    return Y
    
def exp_decay(amp1,amp2,dur1,dur2):
    Y=exp_growth(amp2,amp1,dur1,dur2)
    Y=np.flipud(Y) # used to be flip, but that was not supported by older versions of numpy
    return Y
    
def smooth_it(Y,t):
    Z=np.zeros(Y.size)
    for j in range(-t,t+1):
        Z=Z+np.roll(Y,j)
    return Z

def synthesize_single_waveform(*,N=800,durations=[200,10,30,200],amps=[0.5,10,-1,0]):

    durations=np.array(durations).ravel()
    if (np.sum(durations)>=N-2):
        durations[-1]=N-2-np.sum(durations[0:durations.size-1])
        
    amps=np.array(amps).ravel()
    
    timepoints=np.round(np.hstack((0,np.cumsum(durations)-1))).astype('int');
    
    t=np.r_[0:np.sum(durations)+1]
    
    Y=np.zeros(len(t))
    Y[timepoints[0]:timepoints[1]+1]=exp_growth(0,amps[0],timepoints[1]+1-timepoints[0],durations[0]/4)
    Y[timepoints[1]:timepoints[2]+1]=exp_growth(amps[0],amps[1],timepoints[2]+1-timepoints[1],durations[1])
    Y[timepoints[2]:timepoints[3]+1]=exp_decay(amps[1],amps[2],timepoints[3]+1-timepoints[2],durations[2]/4)
    Y[timepoints[3]:timepoints[4]+1]=exp_decay(amps[2],amps[3],timepoints[4]+1-timepoints[3],durations[3]/5)
    Y=smooth_it(Y,3)
    Y=Y-np.linspace(Y[0],Y[-1],len(t))
    Y=np.hstack((Y,np.zeros(N-len(t))))
    Nmid=int(np.floor(N/2))
    peakind=np.argmax(np.abs(Y))
    Y=np.roll(Y,Nmid-peakind)
    
    return Y
    


#Y=smooth_it(Y,3);
#Y=Y-linspace(Y(1),Y(end),length(Y));
#
#Y=[Y,zeros(1,N-length(Y))];
#
#Nmid=floor(N/2);
#[~,peakind]=max(abs(Y));
#Y=circshift(Y,[0,Nmid-peakind]);
#
#end
#
#function test_synth_waveform
#Y=synthesize_single_waveform(800);
#figure; plot(Y);
#end
#
#function Y=exp_growth(amp1,amp2,dur1,dur2)
#t=1:dur1;
#Y=exp(t/dur2);
#% Want Y(1)=amp1
#% Want Y(end)=amp2
#Y=Y/(Y(end)-Y(1))*(amp2-amp1);
#Y=Y-Y(1)+amp1;
#end
#
#function Y=exp_decay(amp1,amp2,dur1,dur2)
#Y=exp_growth(amp2,amp1,dur1,dur2);
#Y=Y(end:-1:1);
#end
#
#function Z=smooth_it(Y,t)
#Z=Y;
#Z(1+t:end-t)=0;
#for j=-t:t
#    Z(1+t:end-t)=Z(1+t:end-t)+Y(1+t+j:end-t+j)/(2*t+1);
#end;
#end
    
if __name__ == '__main__':
    Y=synthesize_single_waveform()
    import matplotlib.pyplot as plt
    plt.plot(Y)


import numpy as np
from synthesize_single_waveform import synthesize_single_waveform

from mountainlab_pytools import mdaio

processor_name='ephys.synthesize_random_waveforms'
processor_version='0.1'
def synthesize_random_waveforms(*,waveforms_out=None,geometry_out=None,M=5,T=500,K=20,upsamplefac=13,timeshift_factor=3,average_peak_amplitude=10):
    """
    Synthesize random waveforms for use in creating a synthetic timeseries dataset

    Parameters
    ----------
    waveforms_out : OUTPUT
        Path to waveforms mda file. Mx(T*upsamplefac)xK
    geometry_out : OUTPUT
        (Optional) Path to geometry csv file
    M : int
        (Optional) Number of channels
    T : int
        (Optional) Number of timepoints for a waveform, before upsampling
    K : int
        (Optional) Number of waveforms to synthesize
    timeshift_factor : int
        (Optional) Controls amount of timeshift between waveforms on different channels for each template
    upsamplefac : int
        (Optional) used for upsampling the waveforms to avoid discretization artifacts
    average_peak_amplitude : float
        (Optional) used to scale the peak spike amplitude 
    """    
    geometry=None
    avg_durations=[200,10,30,200]
    avg_amps=[0.5,10,-1,0]
    rand_durations_stdev=[10,4,6,20]
    rand_amps_stdev=[0.2,3,0.5,0]
    rand_amp_factor_range=[0.5,1]
    geom_spread_coef1=0.2
    geom_spread_coef2=1
    
    if not geometry:
        geometry=np.zeros((2,M))
        geometry[0,:]=np.arange(1,M+1)
    
    geometry=np.array(geometry)
    avg_durations=np.array(avg_durations)
    avg_amps=np.array(avg_amps)
    rand_durations_stdev=np.array(rand_durations_stdev)
    rand_amps_stdev=np.array(rand_amps_stdev)
    rand_amp_factor_range=np.array(rand_amp_factor_range)
    
    neuron_locations=get_default_neuron_locations(M,K,geometry)

    ## The waveforms_out
    WW=np.zeros((M,T*upsamplefac,K))
    
    for k in range(1,K+1):
        for m in range(1,M+1):
            diff=neuron_locations[:,k-1]-geometry[:,m-1]
            dist=np.sqrt(np.sum(diff**2))
            durations0=np.maximum(np.ones(avg_durations.shape),avg_durations+np.random.randn(1,4)*rand_durations_stdev)*upsamplefac
            amps0=avg_amps+np.random.randn(1,4)*rand_amps_stdev
            waveform0=synthesize_single_waveform(N=T*upsamplefac,durations=durations0,amps=amps0)
            waveform0=np.roll(waveform0,int(timeshift_factor*dist*upsamplefac))
            waveform0=waveform0*np.random.uniform(rand_amp_factor_range[0],rand_amp_factor_range[1])
            WW[m-1,:,k-1]=waveform0/(geom_spread_coef1+dist*geom_spread_coef2)

    peaks=np.max(np.abs(WW),axis=(0,1))
    WW=WW/np.mean(peaks)*average_peak_amplitude

    if waveforms_out:
        mdaio.writemda32(WW,waveforms_out)
        if geometry_out:
            np.savetxt(geometry_out,geometry.transpose(),delimiter=",",fmt="%g")
            return True
        else:
            return True
    else:
        return (WW,geometry)
    
def get_default_neuron_locations(M,K,geometry):
    num_dims=geometry.shape[0]
    neuron_locations=np.zeros((num_dims,K))
    for k in range(1,K+1):
        if K>0:
            ind=(k-1)/(K-1)*(M-1)+1
            ind0=int(ind)
            if ind0==M:
                ind0=M-1
                p=1
            else:
                p=ind-ind0
            if M>0:
                neuron_locations[:,k-1]=(1-p)*geometry[:,ind0-1]+p*geometry[:,ind0]
            else:
                neuron_locations[:,k-1]=geometry[:,0]
        else:
            neuron_locations[:,k-1]=geometry[:,0]
        
    return neuron_locations
    
def test_synthesize_random_waveforms():
    M,T,K = 5,800,20
    upsamplefac=13
    waveforms,geometry = synthesize_random_waveforms(M=M,T=T,K=K,upsamplefac=upsamplefac)
    print(waveforms.shape)
    #waveforms=mdaio.readmda('tmp.waveforms.mda')
    #assert(waveforms.shape==(M,T*upsamplefac,K))
    #import matplotlib.pyplot as plt
    #for k in range(1,K+1):
    #    plt.figure()
    #    for m in range(1,M+1):
    #        plt.plot(waveforms[m-1,:,k-1])
    
    return True

synthesize_random_waveforms.test=test_synthesize_random_waveforms
synthesize_random_waveforms.name = processor_name
synthesize_random_waveforms.version = processor_version

if __name__ == '__main__':
    print ('Running test')
    test_synthesize_random_waveforms()
    



from mountainlab_pytools import mdaio
import numpy as np
from scipy import special
import multiprocessing
import time
import os

class SharedChunkInfo():
    def __init__(self,num_chunks):
        self.timer_timestamp = multiprocessing.Value('d',time.time(),lock=False)
        self.last_appended_chunk = multiprocessing.Value('l',-1,lock=False)
        self.num_chunks=num_chunks
        self.num_completed_chunks = multiprocessing.Value('l',0,lock=False)
        self.lock = multiprocessing.Lock()
    def reportChunkCompleted(self,num):
        with self.lock:
            self.num_completed_chunks.value+=1
    def reportChunkAppended(self,num):
        with self.lock:
            self.last_appended_chunk.value=num
    def lastAppendedChunk(self):
        with self.lock:
            return self.last_appended_chunk.value
    def resetTimer(self):
        with self.lock:
            self.timer_timestamp.value=time.time()
    def elapsedTime(self):
        with self.lock:
            return time.time()-self.timer_timestamp.value
    def printStatus(self):
        with self.lock:
            print('Processed {} of {} chunks...'.format(self.num_completed_chunks.value,self.num_chunks))

def create_filter_kernel(N,samplerate,freq_min,freq_max,freq_wid=1000):
    # Matches ahb's code /matlab/processors/ms_bandpass_filter.m
    # improved ahb, changing tanh to erf, correct -3dB pts  6/14/16    
    T = N / samplerate # total time
    df = 1 / T # frequency grid
    relwid = 3.0; # relative bottom-end roll-off width param, kills low freqs by factor 1e-5.
    
    k_inds=np.arange(0,N)
    k_inds=np.where(k_inds<=(N+1)/2,k_inds,k_inds-N)
    
    fgrid=df*k_inds
    absf=np.abs(fgrid)
    
    val=np.ones(fgrid.shape)
    if freq_min!=0:
        val=val*(1 + special.erf(relwid * (absf - freq_min) / freq_min)) / 2
        val=np.where(np.abs(k_inds)<0.1,0,val) # kill DC part exactly
    if freq_max!=0:
        val=val*(1 - special.erf((absf - freq_max) / freq_wid)) / 2;
    val=np.sqrt(val) # note sqrt of filter func to apply to spectral intensity not ampl
    return val
    
def filter_chunk(num):
    #print('Filtering {}'.format(num))
    opts=g_opts
    #print('Filtering chunk {} of {}'.format(num,opts['num_chunks']))
    in_fname=opts['timeseries'] # The entire (large) input file
    out_fname=opts['timeseries_out'] # The entire (large) output file
    samplerate=opts['samplerate']
    freq_min=opts['freq_min']
    freq_max=opts['freq_max']
    freq_wid=opts['freq_wid']
    chunk_size=opts['chunk_size']
    padding=opts['padding']
    
    X=mdaio.DiskReadMda(in_fname)
    
    chunk_size_with_padding=chunk_size+2*padding
    padded_chunk=np.zeros((X.N1(),chunk_size_with_padding),dtype='float32')    
    
    t1=int(num*opts['chunk_size']) # first timepoint of the chunk
    t2=int(np.minimum(X.N2(),(t1+chunk_size))) # last timepoint of chunk (+1)
    s1=int(np.maximum(0,t1-padding)) # first timepoint including the padding
    s2=int(np.minimum(X.N2(),t2+padding)) # last timepoint (+1) including the padding
    
    # determine aa so that t1-s1+aa = padding
    # so, aa = padding-(t1-s1)
    aa = padding-(t1-s1)
    padded_chunk[:,aa:aa+s2-s1]=X.readChunk(i1=0,N1=X.N1(),i2=s1,N2=s2-s1) # Read the padded chunk
    
    # Do the actual filtering with a DFT with real input
    padded_chunk=np.fft.rfft(padded_chunk) 
    # Subtract off the mean of each channel unless we are doing only a low-pass filter
    if freq_min!=0:
        for m in range(padded_chunk.shape[0]):
            padded_chunk[m,:]=padded_chunk[m,:]-np.mean(padded_chunk[m,:])
    kernel=create_filter_kernel(chunk_size_with_padding,samplerate,freq_min,freq_max,freq_wid)
    kernel=kernel[0:padded_chunk.shape[1]] # because this is the DFT of real data
    padded_chunk=padded_chunk*np.tile(kernel,(padded_chunk.shape[0],1))
    padded_chunk=np.fft.irfft(padded_chunk)
    
    ###########################################################################################
    # Now we wait until we are ready to append to the output file
    # Note that we need to append in order, thus the shared_data object
    ###########################################################################################
    g_shared_data.reportChunkCompleted(num) # Report that we have completed this chunk
    while True: # Alex: maybe there should be a timeout here in case ...
        if num == g_shared_data.lastAppendedChunk()+1:
            break
        time.sleep(0.005) # so we don't saturate the CPU unnecessarily
    
    # Append the filtered chunk (excluding the padding) to the output file
    mdaio.appendmda(padded_chunk[:,padding:padding+(t2-t1)],out_fname)
    
    # Report that we have appended so the next chunk can proceed
    g_shared_data.reportChunkAppended(num)

    # Print status if it has been long enough
    if g_shared_data.elapsedTime()>4:
        g_shared_data.printStatus()
        g_shared_data.resetTimer()
    
def bandpass_filter(*,
        timeseries,timeseries_out,
        samplerate,freq_min,freq_max,freq_wid=1000,
        padding=3000,chunk_size=3000*10,num_processes=os.cpu_count()):
    """
    Apply a bandpass filter to a multi-channel timeseries

    Parameters
    ----------
    timeseries : INPUT
        MxN raw timeseries array (M = #channels, N = #timepoints)
        
    timeseries_out : OUTPUT
        Filtered output (MxN array)
        
    samplerate : float
        The sampling rate in Hz
    freq_min : float
        The lower endpoint of the frequency band (Hz)
    freq_max : float
        The upper endpoint of the frequency band (Hz)
    freq_wid : float
        The optional width of the roll-off (Hz)
    """
    X=mdaio.DiskReadMda(timeseries)
    M=X.N1() # Number of channels
    N=X.N2() # Number of timepoints
    
    num_chunks=int(np.ceil(N/chunk_size))
    print('Chunk size: {}, Padding: {}, Num chunks: {}, Num processes: {}'.format(chunk_size,padding,num_chunks,num_processes))
    
    opts={
        "timeseries":timeseries,
        "timeseries_out":timeseries_out,
        "samplerate":samplerate,
        "freq_min":freq_min,
        "freq_max":freq_max,
        "freq_wid":freq_wid,
        "chunk_size":chunk_size,
        "padding":padding,
        "num_processes":num_processes,
        "num_chunks":num_chunks
    }
    global g_shared_data
    g_shared_data=SharedChunkInfo(num_chunks)
    global g_opts
    g_opts=opts
    mdaio.writemda32(np.zeros([M,0]),timeseries_out)
    
    pool = multiprocessing.Pool(processes=num_processes)
    pool.map(filter_chunk,range(num_chunks),chunksize=1)
    return True
bandpass_filter.name='ephys.bandpass_filter'
bandpass_filter.version='0.1'



from mountainlab_pytools import mdaio
import numpy as np
import multiprocessing
import time
import os

class SharedChunkInfo():
    def __init__(self,num_chunks):
        self.timer_timestamp = multiprocessing.Value('d',time.time(),lock=False)
        self.last_appended_chunk = multiprocessing.Value('l',-1,lock=False)
        self.num_chunks=num_chunks
        self.num_completed_chunks = multiprocessing.Value('l',0,lock=False)
        self.lock = multiprocessing.Lock()
    def reportChunkCompleted(self,num):
        with self.lock:
            self.num_completed_chunks.value+=1
    def reportChunkAppended(self,num):
        with self.lock:
            self.last_appended_chunk.value=num
    def lastAppendedChunk(self):
        with self.lock:
            return self.last_appended_chunk.value
    def resetTimer(self):
        with self.lock:
            self.timer_timestamp.value=time.time()
    def elapsedTime(self):
        with self.lock:
            return time.time()-self.timer_timestamp.value
    def printStatus(self):
        with self.lock:
            print('Processed {} of {} chunks...'.format(self.num_completed_chunks.value,self.num_chunks))

def compute_AAt_matrix_for_chunk(num):
    opts=g_opts
    in_fname=opts['timeseries'] # The entire (large) input file
    out_fname=opts['timeseries_out'] # The entire (large) output file
    chunk_size=opts['chunk_size']
    
    X=mdaio.DiskReadMda(in_fname)
    
    t1=int(num*opts['chunk_size']) # first timepoint of the chunk
    t2=int(np.minimum(X.N2(),(t1+chunk_size))) # last timepoint of chunk (+1)
    
    chunk=X.readChunk(i1=0,N1=X.N1(),i2=t1,N2=t2-t1) # Read the chunk
    
    ret=chunk @ np.transpose(chunk)
    
    return ret

def whiten_chunk(num,W):
    #print('Whitening {}'.format(num))
    opts=g_opts
    #print('Whitening chunk {} of {}'.format(num,opts['num_chunks']))
    in_fname=opts['timeseries'] # The entire (large) input file
    out_fname=opts['timeseries_out'] # The entire (large) output file
    chunk_size=opts['chunk_size']
    
    X=mdaio.DiskReadMda(in_fname)
    
    t1=int(num*opts['chunk_size']) # first timepoint of the chunk
    t2=int(np.minimum(X.N2(),(t1+chunk_size))) # last timepoint of chunk (+1)
    
    chunk=X.readChunk(i1=0,N1=X.N1(),i2=t1,N2=t2-t1) # Read the chunk
    
    chunk=W @ chunk
    
    ###########################################################################################
    # Now we wait until we are ready to append to the output file
    # Note that we need to append in order, thus the shared_data object
    ###########################################################################################
    g_shared_data.reportChunkCompleted(num) # Report that we have completed this chunk
    while True:
        if num == g_shared_data.lastAppendedChunk()+1:
            break
        time.sleep(0.005) # so we don't saturate the CPU unnecessarily
    
    # Append the filtered chunk (excluding the padding) to the output file
    mdaio.appendmda(chunk,out_fname)
    
    # Report that we have appended so the next chunk can proceed
    g_shared_data.reportChunkAppended(num)

    # Print status if it has been long enough
    if g_shared_data.elapsedTime()>4:
        g_shared_data.printStatus()
        g_shared_data.resetTimer()
    
def whiten(*,
        timeseries,timeseries_out,
        chunk_size=30000*10,num_processes=os.cpu_count()
        ):
    """
    Whiten a multi-channel timeseries

    Parameters
    ----------
    timeseries : INPUT
        MxN raw timeseries array (M = #channels, N = #timepoints)
        
    timeseries_out : OUTPUT
        Whitened output (MxN array)

    """
    X=mdaio.DiskReadMda(timeseries)
    M=X.N1() # Number of channels
    N=X.N2() # Number of timepoints

    num_chunks_for_computing_cov_matrix=10
    
    num_chunks=int(np.ceil(N/chunk_size))
    print ('Chunk size: {}, Num chunks: {}, Num processes: {}'.format(chunk_size,num_chunks,num_processes))
    
    opts={
        "timeseries":timeseries,
        "timeseries_out":timeseries_out,
        "chunk_size":chunk_size,
        "num_processes":num_processes,
        "num_chunks":num_chunks
    }
    global g_opts
    g_opts=opts
    
    pool = multiprocessing.Pool(processes=num_processes)
    step=int(np.maximum(1,np.floor(num_chunks/num_chunks_for_computing_cov_matrix)))
    AAt_matrices=pool.map(compute_AAt_matrix_for_chunk,range(0,num_chunks,step),chunksize=1)
    
    AAt=np.zeros((M,M),dtype='float64')
    
    for M0 in AAt_matrices:
        AAt+=M0/(len(AAt_matrices)*chunk_size) ##important: need to fix the denominator here to account for possible smaller chunk
    
    U, S, Ut = np.linalg.svd(AAt, full_matrices=True)
    
    W = (U @ np.diag(1/np.sqrt(S))) @ Ut
    #print ('Whitening matrix:')
    #print (W)
    
    global g_shared_data
    g_shared_data=SharedChunkInfo(num_chunks)
    mdaio.writemda32(np.zeros([M,0]),timeseries_out)
    
    pool = multiprocessing.Pool(processes=num_processes)
    pool.starmap(whiten_chunk,[(num,W) for num in range(0,num_chunks)],chunksize=1)
    
    return True
whiten.name='ephys.whiten'
whiten.version='0.1'



import numpy as np
from mountainlab_pytools import mdaio
import os, shutil

DEBUG=False

def file_extension(fname):
    filename, ext = os.path.splitext(fname)
    return ext

def determine_file_format(ext,dimensions):
    if ext=='.mda':
        return 'mda'
    elif ext=='.npy':
        return 'npy'
    else:
        return 'dat'

def npy_dtype_to_string(dt):
    str=dt.str[1:]
    map={
        "f2":'float16',
        "f4":'float32',
        "f8":'float64',
        "i1":'int8',
        "i2":'int16',
        "i4":'int32',
        "u2":'uint16',
        "u4":'uint32'
    }
    return map[str]

def get_num_bytes_per_entry_from_dt(dt):
    if dt == 'uint8':
        return 1
    if dt == 'float32':
        return 4
    if dt == 'int16':
        return 2
    if dt == 'int32':
        return 4
    if dt == 'uint16':
        return 2
    if dt == 'float64':
        return 8
    if dt == 'uint32':
        return 4
    return None

def copy_raw_file_data(input,output,*,start_byte,num_entries,dtype,dtype_out):
    if dtype != dtype_out:
        raise Exception('Copying from one datatype to another not yet implemented')
    size1=get_num_bytes_per_entry_from_dt(dtype)
    size2=get_num_bytes_per_entry_from_dt(dtype_out)
    bufsize=10000
    with open(input,'rb') as f1:
        f1.seek(start_byte)
        with open(output,'wb') as f2:
            while num_entries:
                chunk_num_entries = min(bufsize,num_entries)
                data = f1.read(chunk_num_entries*size1)
                ## convert data here
                f2.write(data)
                num_entries -= chunk_num_entries

def determine_npy_header_size(fname):
    with open(fname,'rb') as f:
        header_str=f.readline() ## I think it ends with \n
        return f.tell()

processor_name='ephys.convert_array'
processor_version='0.1'
def convert_array(*,input,output,format='',format_out='',dimensions='',dtype='',dtype_out='', channels=''):
    """
    Convert a multi-dimensional array between various formats ('.mda', '.npy', '.dat') based on the file extensions of the input/output files

    Parameters
    ----------
    input : INPUT
        Path of input array file (can be repeated for concatenation).
    output : OUTPUT
        Path of the output array file.
        
    format : string
        The format for the input array (mda, npy, dat), or determined from the file extension if empty
    format_out : string
        The format for the output input array (mda, npy, dat), or determined from the file extension if empty
    dimensions : string
        Comma-separated list of dimensions (shape). If empty, it is auto-determined, if possible, by the input array. If second dim is -1 then it will be extrapolated from file size / first dim.
    dtype : string
        The data format for the input array. Choices: int8, int16, int32, uint16, uint32, float32, float64 (possibly float16 in the future).
    dtype_out : string
        The data format for the output array. If empty, the dtype for the input array is used.
    channels : string
        Comma-seperated list of channels to keep in output. Zero-based indexing. Only works for .dat to .mda conversions.
    """    
    if isinstance(input, (list,)):
        multifile = True
        inputs = input
        input = inputs[0]
    else:
        multifile = False
    
    format_in=format
    if not format_in:
        format_in=determine_file_format(file_extension(input),dimensions)
    if not format_out:
        format_out=determine_file_format(file_extension(output),dimensions)
    print ('Input/output formats: {}/{}'.format(format_in,format_out))
    ext_in = file_extension(input)

    dims=None

    if (format_in=='mda') and (dtype==''):
        header=mdaio.readmda_header(input)
        dtype=header.dt
        dims=header.dims

    if (format_in=='npy') and (dtype==''):
        A=np.load(input,mmap_mode='r')
        dtype=npy_dtype_to_string(A.dtype)
        dims=A.shape
        A=0

    if dimensions:
        dims2=[int(entry) for entry in dimensions.split(',')]
        if dims:
            if len(dims) != len(dims2):
                raise Exception('Inconsistent number of dimensions for input array')
            if not np.all(np.array(dims)==np.array(dims2)):
                raise Exception('Inconsistent dimensions for input array')
        dims=dims2

    if not dtype_out:
        dtype_out=dtype

    if not dtype:
        raise Exception('Unable to determine datatype for input array')

    if not dtype_out:
        raise Exception('Unable to determine datatype for output array')
    
    if (dims[1] == -1) and (dims[0] > 0):
        if ((dtype) and (format_in=='dat') and not multifile):
            bits      = int(dtype[-2:])        # number of bits per entry of dtype, TODO: make this smarter
            filebytes = os.stat(input).st_size # bytes in input file
            entries   = int(filebytes/(int(bits/8))) # entries in input file
            dims[1]   = int(entries/dims[0])   # caclulated second dimension
            if DEBUG:
                print(bits)
                print(filebytes)
                print(int(filebytes/(int(bits/8))))
                print(dims)
        else:
            raise Exception('Could not infer dimensions')

    if not dims:       
        raise Exception('Unable to determine dimensions for input array')

    if not channels:
        channels = range(0, dims[0])
    else:
        channels = np.array([int(entry) for entry in channels.split(',')])

    if DEBUG:
        print(channels);

    print ('Using dtype={}, dtype_out={}, dimensions={}'.format(dtype,dtype_out,','.join(str(item) for item in dims)))
    if (format_in==format_out) and ((dtype==dtype_out) or (dtype_out=='')):
        if multifile and (format_in == 'dat'):
            print('Concatenating Files')
            with open(output, "wb") as outfile:
                for input_file in inputs:
                    with open(input_file, "rb") as inpt:
                        outfile.write(inpt.read())
        elif not multifile:
            print ('Simply copying file...')
            shutil.copyfile(input,output)
            print ('Done.')
        return True

    if format_out=='dat' and not multifile:
        if format_in=='mda':
            H=mdaio.readmda_header(input)
            copy_raw_file_data(input,output,start_byte=H.header_size,num_entries=np.product(dims),dtype=dtype,dtype_out=dtype_out)
            return True
        elif format_in=='npy':
            print ('Warning: loading entire array into memory. This should be avoided in the future.')
            A=np.load(input,mmap_mode='r').astype(dtype=dtype_out,order='F',copy=False)
            A=A.ravel(order='F')
            A.tofile(output)
            # The following was problematic because of row-major ordering, i think
            #header_size=determine_npy_header_size(input)
            #copy_raw_file_data(input,output,start_byte=header_size,num_entries=np.product(dims),dtype=dtype,dtype_out=dtype_out)
            return True
        elif format_in=='dat':
            raise Exception('This case not yet implemented.')
        else:
            raise Exception('Unexpected case.')

    elif (format_out=='mda') or (format_out=='npy'):
        if format_in=='dat' and multifile:
            print ('Warning: loading entire array into memory. This should be avoided in the future.')
            A=np.fromfile(inputs[0],dtype=dtype,count=np.product(dims))
            A=A.reshape(tuple(dims),order='F')
            A=A[channels,:] 
            for inputn in inputs[1:]:
                An=np.fromfile(inputn,dtype=dtype,count=np.product(dims))
                An=An.reshape(tuple(dims),order='F')
                An=An[channels,:]
                A=np.concatenate((A,An),axis=0) 
            if format_out=='mda':
                mdaio.writemda(A,output,dtype=dtype_out)
            else:
                mdaio.writenpy(A,output,dtype=dtype_out)
            return True
        elif format_in=='dat' and not multifile:
            print ('Warning: loading entire array into memory. This should be avoided in the future.')
            A=np.fromfile(input,dtype=dtype,count=np.product(dims))
            A=A.reshape(tuple(dims),order='F')
            A=A[channels,:]
            if format_out=='mda':
                mdaio.writemda(A,output,dtype=dtype_out)
            else:
                mdaio.writenpy(A,output,dtype=dtype_out)
            return True
        elif format_in=='mda' and not multifile:
            print ('Warning: loading entire array into memory. This should be avoided in the future.')
            A=mdaio.readmda(input)
            if format_out=='mda':
                mdaio.writemda(A,output,dtype=dtype_out)
            else:
                mdaio.writenpy(A,output,dtype=dtype_out)
            return True
        elif format_in=='mda' and multifile:
            print ('Warning: loading entire array into memory. This should be avoided in the future.')
            A=mdaio.readmda(inputs[0])
            A=A[channels,:] 
            for inputn in inputs[1:]:
                An=mdaio.readmda(inputn)
                An=An[channels,:]
                A=np.concatenate((A,An),axis=0) 
            if format_out=='mda':
                mdaio.writemda(A,output,dtype=dtype_out)
            else:
                mdaio.writenpy(A,output,dtype=dtype_out)
            return True
        elif format_in=='npy' and not multifile:
            print ('Warning: loading entire array into memory. This should be avoided in the future.')
            A=np.load(input,mmap_mode='r').astype(dtype=dtype,order='F',copy=False)
            if format_out=='mda':
                mdaio.writemda(A,output,dtype=dtype_out)
            else:
                mdaio.writenpy(A,output,dtype=dtype_out)
            return True
        elif format_in=='npy' and multifile:
            print ('Warning: loading entire array into memory. This should be avoided in the future.')
            A=np.load(inputs[0],mmap_mode='r').astype(dtype=dtype,order='F',copy=False)
            A=A[channels,:] 
            for inputn in inputs[1:]:
                An=np.load(inputn,mmap_mode='r').astype(dtype=dtype,order='F',copy=False)
                An=An[channels,:]
                A=np.concatenate((A,An),axis=0) 
            if format_out=='mda':
                mdaio.writemda(A,output,dtype=dtype_out)
            else:
                mdaio.writenpy(A,output,dtype=dtype_out)
            return True
        else:
            raise Exception('Unexpected case.')
    else:
        raise Exception('Unexpected output format: {}'.format(format_out))

    raise Exception('Unexpected error.')

convert_array.name=processor_name
convert_array.version=processor_version

def test_convert_array(dtype='int32',shape=[12,3,7]):
    X=np.array(np.random.normal(0,1,shape),dtype=dtype)
    np.save('test_convert1.npy',X)
    convert_array(input='test_convert1.npy',output='test_convert2.mda') # npy -> mda
    convert_array(input='test_convert2.mda',output='test_convert3.npy') # mda -> npy
    convert_array(input='test_convert3.npy',output='test_convert4.dat') # npy -> dat
    convert_array(input='test_convert4.dat',output='test_convert5.npy',dtype=dtype,dimensions=','.join(str(entry) for entry in X.shape))  # dat -> npy
    convert_array(input='test_convert5.npy',output='test_convert6.mda') # npy -> mda
    convert_array(input='test_convert6.mda',output='test_convert7.dat') # mda -> dat
    convert_array(input='test_convert7.dat',output='test_convert8.mda',dtype=dtype,dimensions=','.join(str(entry) for entry in X.shape)) # dat -> mda
    Y=mdaio.readmda('test_convert8.mda')
    print(np.max(np.abs(X-Y)),Y.dtype)



from mountainlab_pytools import mdaio
import time

class TimeseriesChunkInfo:
    def __init__(self):
        self.t1a=0
        self.t2a=0
        self.t1=0
        self.t2=0
        self.size=0

class TimeseriesChunkReader:
    def __init__(self, chunk_size=0, chunk_size_mb=0, overlap_size=0, t1=-1, t2=-1, verbose=True):
        # Note that the actual chunk size will be the maximum of chunk_size,overlap_size and chunk_size_mb*1e6/(M*4)
        self._chunk_size=chunk_size
        self._chunk_size_mb=chunk_size_mb
        self._overlap_size=overlap_size
        self._t1=t1
        self._t2=t2
        self._elapsed_reading=0
        self._elapsed_running=0
        self._verbose=verbose
    def run(self, mdafile_path_or_diskreadmda, func):
        if (type(mdafile_path_or_diskreadmda)==str):
            X=mdaio.DiskReadMda(mdafile_path_or_diskreadmda)
        else:
            X=mdafile_path_or_diskreadmda
        M,N = X.N1(),X.N2()
        cs=max([self._chunk_size,int(self._chunk_size_mb*1e6/(M*4)),M])        
        if self._t1<0:
            self._t1=0
        if self._t2<0:
            self._t2=N-1
        t=self._t1
        while t <= self._t2:
            t1=t
            t2=min(self._t2,t+cs-1)
            s1=max(0,t1-self._overlap_size)
            s2=min(N-1,t2+self._overlap_size)
            
            timer=time.time()
            chunk=X.readChunk(i1=0, N1=M, i2=s1, N2=s2-s1+1)
            self._elapsed_reading+=time.time()-timer
            
            info=TimeseriesChunkInfo()
            info.t1=t1
            info.t2=t2
            info.t1a=t1-s1
            info.t2a=t2-s1
            info.size=t2-t1+1
            
            timer=time.time()
            if not func(chunk, info):
                return False
            self._elapsed_running+=time.time()-timer                
                
            t=t+cs
        if self._verbose:
            print('Elapsed for TimeseriesChunkReader: %g sec reading, %g sec running' % (self._elapsed_reading,self._elapsed_running))
        return True
    def elapsedReading(self):
        return self._elapsed_reading
    def elapsedRunning(self):
        return self._elapsed_running



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

pkgs = setuptools.find_packages()
print('found these packages:', pkgs)

setuptools.setup(
    name="ml_ephys",
    version="0.0.1",
    author="Jeremy Magland",
    author_email="",
    description="ephys tools for MountainLab",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/magland/ml_ephys",
    packages=pkgs,
    package_data={
        # Include all processor files
        '': ['*.mp']
    },
    install_requires=
    [
        'mountainlab_pytools',
        'deepdish',
        'scipy',
        'numpy',
        'numpydoc',
        'h5py'
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
)



import os
import sys

from mountainlab_pytools import processormanager as pm

import p_synthesize_timeseries
import p_synthesize_random_waveforms
import p_synthesize_random_firings

#import p_synthesize_drifting_timeseries

PM=pm.ProcessorManager()

PM.registerProcessor(p_synthesize_timeseries.synthesize_timeseries)
PM.registerProcessor(p_synthesize_random_waveforms.synthesize_random_waveforms)
PM.registerProcessor(p_synthesize_random_firings.synthesize_random_firings)

#PM.registerProcessor(p_synthesize_drifting_timeseries.synthesize_drifting_timeseries)

if not PM.run(sys.argv):
    exit(-1)
